/*****************************************************************************
 *
 * Copyright (C) 2023 Daniel Ramirez <dr@igh.de>
 *               2023 Florian Pose <fp@igh.de>
 *
 * This file is part of the QtPdWidgets library.
 *
 * The QtPdWidgets library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * The QtPdWidgets library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the QtPdWidgets Library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include "Parameter.h"

#include "QtPdWidgets2/ParameterSetWidget.h"

#include <pdcom5/Subscriber.h>
#include <pdcom5/Subscription.h>

#include <QDebug>
#include <QJsonArray>
#include <QJsonObject>

using Pd::Parameter;

/*****************************************************************************
 * public
 ****************************************************************************/

Parameter::Parameter(Pd::ParameterSetWidget *widget) :
    PdCom::Subscriber(PdCom::event_mode),
    widget(widget), 
    scale {1.0},
    offset {0.0},
    dataPresent(false), 
    savedDataPresent(false)
{}

/****************************************************************************/

Parameter::~Parameter()
{}

/****************************************************************************/

void Parameter::fromJson(const QJsonValue &value)
{
    QJsonObject obj(value.toObject());

    url = obj["url"].toString();
    connectionPath =
        url.adjusted(QUrl::RemoveAuthority | QUrl::RemoveScheme).toString();

    double scale(1.0);
    if (obj.contains("scale")) {
        auto scaleValue(obj["scale"]);
        if (not scaleValue.isDouble()) {
            throw(std::invalid_argument(
                    "scale must be a scalar double value."));
        }

        scale = obj["scale"].toDouble();
        if (scale == 0.0) {
            throw(std::invalid_argument("scale must not be zero"));
        }
    }

    if (obj.contains("value")) {
        auto value(obj["value"]);
        if (value.isArray()) {
            int nelem = value.toArray().count();
            for (int k = 0; k < nelem; k++) {
                savedItems.append(value.toArray().at(k).toDouble() / scale);
            }
            savedDataPresent = true;
        }
        else if (value.isDouble()) {
            savedItems.append(value.toDouble() / scale);
            savedDataPresent = true;
        }
        else {
            throw(std::invalid_argument("Value type not yet implemented."));
            savedDataPresent = false;
        }
    }
    else {
        savedItems = QVariantList();
        savedDataPresent = false;
    }    
        
#ifdef DEBUG_PD_PARAMETER
    qWarning() << __func__ << "connection path" << connectionPath;
    qWarning() << __func__ << "saved value" << savedItems;
#endif
}

/****************************************************************************/

void Parameter::replaceUrl(const QUrl &oldUrl, const QUrl &newUrl)
{
    QUrl dataSourceUrl(url.adjusted(
            QUrl::RemovePassword | QUrl::RemovePath | QUrl::RemoveQuery
            | QUrl::RemoveFragment));
    if (dataSourceUrl != oldUrl) {
        return;
    }

    url.setScheme(newUrl.scheme());
    url.setAuthority(newUrl.authority());  // user, pass, host, port
}

/****************************************************************************/

void Parameter::connectParameter(QtPdCom::Process *p)
{
    if (p) {
        pv = p->find(connectionPath);
        
        // Subscribe to process variable as ScalarSubscriber
        subscription = PdCom::Subscription(*this, 
            p->find(connectionPath));

        try {
            subscription.poll();
        }
        catch (PdCom::Exception &e) {
            QString msg(tr("Parameter poll failed: %1").arg(e.what()));
            qWarning() << msg;
        }
#ifdef DEBUG_PD_PARAMETER
        qDebug() << __func__ << url;
        qDebug() << __func__ << p;
        qDebug() << __func__ << &pv;
        qDebug() << __func__ << &subscription;
#endif
    }
}

/***************************************************************************/

void Parameter::setCurrentValue(const QVariant &value)
{
    if (pv.empty() or not dataPresent) {
        return;
    }

    auto nelem(pv.getSizeInfo().totalElements());
    int cnt = nelem;
    bool isNumber {false};
    value.toDouble(&isNumber);
    bool isList = value.canConvert<QVariantList>();  // see assistant about
                                                     // canConvert() and
                                                     // convert()

    // a string can be stored in an uint8 array as well
    if (value.canConvert<QString>() and not isNumber and not isList) {
        if (pv.getTypeInfo().type == PdCom::TypeInfo::uint8_T) {
            QString s(value.toString());
            QByteArray data;
            do { /* the string with the trailing 0 must not be longer than the
                    array we need to chop the string and not the byteArray
                    because there can be an utf8 char at the end which is
                    longer than one byte */
                data = s.toUtf8();
                s.chop(1);
            } while (data.size() > cnt - 1);
            pv.setValue(
                    data.data(),
                    PdCom::details::TypeInfoTraits<uint8_t>::type_info.type,
                    data.size());
            emit dataChanged();
        }
        qWarning() << "type error: string is supplied to Parameter "
                   << "but variable is not uint8";
        return;  // can't be anything else
    }

    if (isList) {
        cnt = qMin(value.toList().count(), (int) nelem);
    }

    PdCom::TypeInfo::DataType type = pv.getTypeInfo().type;

    switch (type) {
        case PdCom::TypeInfo::boolean_T:
            qWarning() << __func__ << __LINE__;
            qWarning() << type << "not implemented";
            break;
        case PdCom::TypeInfo::uint8_T:
            qWarning() << __func__ << __LINE__;
            qWarning() << type << "not implemented";
            break;
        case PdCom::TypeInfo::uint16_T:
            qWarning() << __func__ << __LINE__;
            qWarning() << type << "not implemented";
            break;
        case PdCom::TypeInfo::uint32_T: {
            qWarning() << __func__ << __LINE__;
            uint32_t data[nelem];
            if (isList) {
                for (int i = 0; i < cnt; i++) {
                    data[i] = scaleToProcess(
                            value.toList().at(i).toULongLong(&isNumber));
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            else { /* it might be a scalar, then set all items to the
                        same value */
                for (int i = 0; i < cnt; i++) {
                    data[i] = scaleToProcess(value.toULongLong(&isNumber));
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            pv.setValue(
                    data,
                    PdCom::details::TypeInfoTraits<uint32_t>::type_info.type,
                    cnt);
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::uint64_T: {
            qWarning() << __func__ << __LINE__;
            uint64_t data[nelem];
            if (isList) {
                for (int i = 0; i < cnt; i++) {
                    data[i] = scaleToProcess(
                            value.toList().at(i).toULongLong(&isNumber));
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            else { /* it might be a scalar, then set all items to the
                        same value */
                for (int i = 0; i < cnt; i++) {
                    data[i] = scaleToProcess(value.toULongLong(&isNumber));
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            pv.setValue(
                    data,
                    PdCom::details::TypeInfoTraits<uint64_t>::type_info.type,
                    cnt);
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::int8_T:
            qWarning() << type << "not implemented";
            break;
        case PdCom::TypeInfo::int16_T:
            qWarning() << type << "not implemented";
            break;
        case PdCom::TypeInfo::int32_T: {
            int32_t data[nelem];
            if (isList) {
                for (int i = 0; i < cnt; i++) {
                    data[i] = scaleToProcess(
                            value.toList().at(i).toLongLong(&isNumber));
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            else { /* it might be a scalar, then set all items to the
                        same value */
                for (int i = 0; i < cnt; i++) {
                    data[i] = scaleToProcess(value.toLongLong(&isNumber));
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            pv.setValue(
                    data,
                    PdCom::details::TypeInfoTraits<int32_t>::type_info.type,
                    cnt);
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::int64_T: {
            int64_t data[nelem];
            if (isList) {
                for (int i = 0; i < cnt; i++) {
                    data[i] = value.toList().at(i).toLongLong(&isNumber);
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            else {
                for (int i = 0; i < cnt; i++) {
                    data[i] = value.toLongLong(&isNumber);
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            pv.setValue(
                    data,
                    PdCom::details::TypeInfoTraits<int64_t>::type_info.type,
                    cnt);
            emit dataChanged();
        } break;

        case PdCom::TypeInfo::single_T:
            qWarning() << type << "not implemented";
            break;

        case PdCom::TypeInfo::double_T: {
            double data[cnt];
            if (isList) {
                for (int i = 0; i < cnt; i++) {
                    data[i] = scaleToProcess(
                            value.toList().at(i).toDouble(&isNumber));
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            else {
                for (int i = 0; i < cnt; i++) {
                    data[i] = scaleToProcess(value.toDouble(&isNumber));
                    if (not isNumber) {
                        qWarning() << "type error: list is supplied to "
                                      "Parameter but "
                                   << "element are not numeric";
                    }
                }
            }
            pv.setValue(
                    data,
                    PdCom::details::TypeInfoTraits<double>::type_info.type,
                    cnt);
            emit dataChanged();
        } break;

        default:
            qWarning() << "unknown datatype";
            break;
    }
}

/*****************************************************************************
 * private
 ****************************************************************************/

void Parameter::newValues(std::chrono::nanoseconds)
{
    auto pv(getVariable());
    auto nelem(pv->getSizeInfo().totalElements());
#ifdef DEBUG_PD_PARAMETER
    qDebug() << __func__ << url;
    qDebug() << __func__ << &pv;
    qDebug() << __func__ << "pv empty" << pv->empty();
    qDebug() << __func__ << "current items" << currentItems;
#endif

    currentItems.clear();
    PdCom::TypeInfo::DataType type = pv->getTypeInfo().type;

    switch (type) {
        case PdCom::TypeInfo::boolean_T: {
            bool boolean[nelem];
            PdCom::details::copyData(
                    boolean,
                    PdCom::details::TypeInfoTraits<uint64_t>::type_info.type,
                    subscription.getData(),
                    pv->getTypeInfo().type,
                    nelem);
            for (size_t i = 0; i < nelem; i++) {
                currentItems.append((bool) boolean[i]);
            }
            dataPresent = true;
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::uint8_T: {
            uint8_t uInt[nelem];
            PdCom::details::copyData(
                    uInt,
                    PdCom::details::TypeInfoTraits<uint64_t>::type_info.type,
                    subscription.getData(),
                    pv->getTypeInfo().type,
                    nelem);
            for (size_t i = 0; i < nelem; i++) {
                currentItems.append((qint8) uInt[i]);
            }
            dataPresent = true;
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::uint16_T:
            qWarning() << __func__ << __LINE__;
            qWarning() << type << "not implemented";
            break;
        case PdCom::TypeInfo::uint32_T: {
            uint32_t uInt[nelem];
            PdCom::details::copyData(
                    uInt,
                    PdCom::details::TypeInfoTraits<uint64_t>::type_info.type,
                    subscription.getData(),
                    pv->getTypeInfo().type,
                    nelem);
            for (size_t i = 0; i < nelem; i++) {
                currentItems.append((qint32) uInt[i]);
            }
            dataPresent = true;
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::uint64_T: {
            uint64_t uInt[nelem];
            PdCom::details::copyData(
                    uInt,
                    PdCom::details::TypeInfoTraits<uint64_t>::type_info.type,
                    subscription.getData(),
                    pv->getTypeInfo().type,
                    nelem);
            for (size_t i = 0; i < nelem; i++) {
                currentItems.append((qint64) uInt[i]);
            }
            dataPresent = true;
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::int8_T: {
            int8_t sInt[nelem];
            PdCom::details::copyData(
                    sInt,
                    PdCom::details::TypeInfoTraits<uint64_t>::type_info.type,
                    subscription.getData(),
                    pv->getTypeInfo().type,
                    nelem);
            for (size_t i = 0; i < nelem; i++) {
                currentItems.append((qint8) sInt[i]);
            }
            dataPresent = true;
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::int16_T:
            qWarning() << __func__ << __LINE__;
            qWarning() << type << "not implemented";
            break;
        case PdCom::TypeInfo::int32_T: {
            int32_t sInt[nelem];
            PdCom::details::copyData(
                    sInt,
                    PdCom::details::TypeInfoTraits<int64_t>::type_info.type,
                    subscription.getData(),
                    pv->getTypeInfo().type,
                    nelem);
            for (size_t i = 0; i < nelem; i++) {
                currentItems.append((qint32) sInt[i]);
            }
            dataPresent = true;
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::int64_T: {
            int64_t sInt[nelem];
            PdCom::details::copyData(
                    sInt,
                    PdCom::details::TypeInfoTraits<int64_t>::type_info.type,
                    subscription.getData(),
                    pv->getTypeInfo().type,
                    nelem);
            for (size_t i = 0; i < nelem; i++) {
                currentItems.append((qint64) sInt[i]);
            }
            dataPresent = true;
            emit dataChanged();
        } break;
        case PdCom::TypeInfo::single_T:
            qWarning() << __func__ << __LINE__;
            qWarning() << type << "not implemented";
            break;
        case PdCom::TypeInfo::double_T: {
            double d[nelem];
            PdCom::details::copyData(
                    d,
                    PdCom::details::TypeInfoTraits<double>::type_info.type,
                    subscription.getData(),
                    pv->getTypeInfo().type,
                    nelem);

            for (size_t i = 0; i < nelem; i++) {
                currentItems.append(
                        scaleToUser(d[i]));  // FIXME: scale parameter
            }
            dataPresent = true;
            emit dataChanged();
        } break;
        default:
            qWarning() << __func__ << __LINE__;
            qWarning() << "unknown datatype";
            break;
    }
}

/****************************************************************************/

void Parameter::stateChanged(const PdCom::Subscription &sub)
{
#ifdef DEBUG_PD_PARAMETER
    qDebug() << __func__ << url << (int) sub.getState();
#endif
    if (sub.getState() != PdCom::Subscription::State::Active) {
        dataPresent = false;
        emit dataChanged();
    }
}

/****************************************************************************/
